#ifndef _ViewCellBsp_H__
#define _ViewCellBsp_H__

#include "Mesh.h"
class ViewCell;
class Plane3;
//class Mesh;

//namespace GtpVisibilityPreprocessor {
  
class BspInterior;


  /**
     BspNode abstract class serving for interior and leaf node implementation
  */
  class BspNode 
  {
  public:
	  /** Determines whether this node is a leaf or not
	  @return true if leaf
	  */
	  virtual bool IsLeaf() const = 0;
    
	  /** Determines whether this node is a root
	  @return true if root
	  */
	  virtual bool IsRoot() const;

  protected:

	  /// parent of this node
	  BspInterior *mParent;
  };

  /** BSP interior node implementation */
  class BspInterior : public BspNode 
  {   
  public:
	  BspInterior(Plane3 plane): mPlane(plane) {}
	  /** @return false since it is an interior node */
	  bool IsLeaf() const;

	  BspNode *GetBack() {return mBack;}
	  BspNode *GetFront() {return mFront;}

	  Plane3 *GetPlane();

	  void ReplaceChildLink(BspNode *oldChild, BspNode *newChild);
	  void SetupChildLinks(BspNode *b, BspNode *f);

  protected:
		
	  /// Splitting plane corresponding to this node
	  Plane3 mPlane;
	  /// back node
	  BspNode *mBack;
	  /// front node
	  BspNode *mFront;
  };
  
  
  /** BSP leaf node implementation */
  class BspLeaf : public BspNode 
  {
  public:
	  BspLeaf(ViewCell *viewCell = NULL);
    
	  /** @return true since it is an interior node */
	  bool IsLeaf() const;
	  ViewCell *GetViewCell();

  protected:
   
	  /// polygonal representation of this viewcell
	  /// if NULL this note does not correspond to feasible viewcell
	  ViewCell *mViewCell;
  };
  
  /** Implementation of the ViewCell BSP tree */
  class BspTree 
  {
  public:
	  struct BspTraversalData
	  {  
		  BspNode *mNode;
		  BspInterior *mParent;

		  Mesh mViewCell;
		  int mDepth;
		      
		  BspTraversalData() {}
		  BspTraversalData(BspNode *n, BspInterior *p, const Mesh &m, const int d): 
		  mNode(n), mParent(p), mViewCell(m), mDepth(d) {}
      };
	 
	  /** Constructor takes a pointer to the cell corresponding to the whole
	  viewspace */
	  BspTree(ViewCell *cell);
        
  protected:
	  /** Selects a splitting plane from the given polygons. */
	  Plane3 SelectPlane(Mesh *polys);

	  void Subdivide();
	   
	  BspNode *SubdivideNode(BspLeaf *leaf, BspInterior *parent, 
							 Mesh *viewCell, const int depth, 
							 Mesh &frontPolys, Mesh &backPolys);

	  /// Pointer to the root of the tree
	  BspNode *mRoot;
	  /// Pointer to the root cell of the viewspace
	  ViewCell *mRootCell;
  };
  
//}; // GtpVisibilityPreprocessor

#endif
